#!/bin/sh

NO_REBOOT="0"
CHECK_VERSION="0"
FILE_LOCK_UPGRADE=/tmp/lock_multiple_upgrade
BOOT_VERSION_FILE=/tmp/boot_version_file
BANK_ACTIVE=
BANK_NEXT=
BANK_1_TRY=
BANK_2_TRY=
CUR_STAGE=0
BOOTLOADER_IS_FOUND=0
FIRMWARE_IS_FOUND=0
BOOTLOADER_INSTALL_REQUIRED=0
FIRMWARE_INSTALL_REQUIRED=0

DEVICE_BOOT=/dev/mtd0
DEVICE_ENV=/dev/mtd1
BLOCK_SIZE=0x20000
ENV1_OFFSET=0
ENV2_OFFSET=0x40000
ENV_SIZE_IN_BLOCKS=2
FIP1_OFFSET=0x0
FIP2_OFFSET=0x100000
FIP_SIZE_IN_BLOCKS=8

fip_file="fip"
env_file="u-boot-env"
dtb_file="dtb"
kernel_file="kernel"
rootfs_file="rootfs"

VOLUME_NAME__DTB="ubi_dtb"
VOLUME_NAME__KERNEL="ubi_kernel"
VOLUME_NAME__ROOTFS="ubi_rootfs"

firmware_mtd_list="$dtb_file $kernel_file $rootfs_file"
bootloader_mtd_list="$fip_file $env_file"

# $1 - exit code
process_exit() {
  code=$1

  if [ "$code" -ne "0" ]; then
    rm ${FILE_LOCK_UPGRADE}
    echo "Lock file removed" > /dev/console
  fi

  exit $code
}

read_env_variables() {
  echo "Reading variables..." > /dev/console

  #test if cur_bootbank var exist
  ubootenv -g "cur_bootbank"

  #if we have old env params
  if [[ "$?" == "0" ]]; then
    BANK_ACTIVE=$(($(ubootenv -g "cur_bootbank" | cut -d '=' -f2)))
  else
    BANK_ACTIVE=$(($(ubootenv -g "bank_active" | cut -d '=' -f2)))
    BANK_1_TRY=$(($(ubootenv -g "bank1_try" | cut -d '=' -f2)))
    BANK_2_TRY=$(($(ubootenv -g "bank2_try" | cut -d '=' -f2)))
  fi
  
  echo "  bank_active is ${BANK_ACTIVE}" > /dev/console
  echo "  bank1_try is ${BANK_1_TRY}" > /dev/console
  echo "  bank2_try is ${BANK_2_TRY}" > /dev/console
}

save_env_variables() {
  echo "Saving variables..." > /dev/console

  ubootenv -s bank_active ${BANK_ACTIVE}
  ubootenv -s bank1_try ${BANK_1_TRY}
  ubootenv -s bank2_try ${BANK_2_TRY}
}

check_md5_match() {
  file_base=$1

  md5sum1=$(awk '{ print $1 }' < "./$file_base.md5sum")
  md5sum2=$(md5sum "./$file_base.bin" | awk '{ print $1 }')

  if [ "$md5sum1" != "$md5sum2" ]; then
    echo "[Error] md5sum's isn't equal!" > /dev/console
    process_exit 1
  fi
}

# $1 arg == mtd
# $2 arg == action (0 == write, 1 == erase)
upgrade_failed() {
  CUR_STAGE=$(($CUR_STAGE + 1))
  MTD=$1
  ACTION=$2

  if [[ "$ACTION" == "0" ]]
  then
    echo "  Stage ${CUR_STAGE}: writing ${MTD} has been FAILED" > /dev/console
  else
    echo "  Stage ${CUR_STAGE}: erasing ${MTD} has been FAILED" > /dev/console
  fi

  echo "[Error] Firmware upgrade has been FAILED. Please check files or permissions!" > /dev/console
  process_exit 1
}

# $1 arg == mtd
# $2 action (0 == write, 1 == erase)
log_upgrade_ok() {
  CUR_STAGE=$(($CUR_STAGE + 1))
  MTD=$1
  ACTION=$2

  if [[ "$ACTION" == "0" ]]
  then
    echo "  Stage ${CUR_STAGE}: writing ${MTD} has been successfully completed" > /dev/console
  else
    echo "  Stage ${CUR_STAGE}: erasing ${MTD} has been successfully completed" > /dev/console
  fi
}

# $1 arg == mtd
# $2 arg == BLOCK_COUNT
erase_flash() {
  MTD=$1
  BLOCK_COUNT=$2

  flash_erase "${MTD}" 0 "${BLOCK_COUNT}" > /dev/null

  if [ "$?" -ne "0" ]; then
    upgrade_failed ${MTD} 1
  else
    log_upgrade_ok ${MTD} 1
  fi
}

# $1 arg == mtd
# $2 arg == path to file
write_flash() {
  MTD=$1
  WRITE_PATH=$2

  nandwrite -pk "${MTD}" "${WRITE_PATH}" > /dev/null

  if [ "$?" -ne "0" ]; then
    upgrade_failed ${MTD} 0
  else
    log_upgrade_ok ${MTD} 0
  fi
}

# $1 arg == device num
# $2 arg == volume name
get_vol_num_from_vol_name() {
	dev_num="$1"
	vol_name="$2"
	info=$(ubinfo -d "${dev_num}" -N "${vol_name}")
	set $info
	echo "/dev/ubi${dev_num}_$3"
}

check_files_exist() {
  for i in $bootloader_mtd_list; do
    if [ -f "./$i.bin" ]; then
      if [ ! -f "./$i.md5sum" ]; then
        echo "[Error] ./$i.md5sum doesn't exist!" > /dev/console
        process_exit 1
      fi
      check_md5_match "$i"

      if [ "$i" == "$fip_file" ] || [ "$i" == "$env_file" ]; then
        BOOTLOADER_IS_FOUND=1
        BOOTLOADER_INSTALL_REQUIRED=1
      fi
    fi
  done

  for i in $firmware_mtd_list; do
    if [ -f "./$i.bin" ]; then
      if [ ! -f "./$i.md5sum" ]; then
        echo "[Error] ./$i.md5sum doesn't exist!" > /dev/console
        process_exit 1
      fi
      check_md5_match "$i"

      if [ "$i" == "$dtb_file" ] || [ "$i" == "$kernel_file" ] || [ "$i" == "$rootfs_file" ]; then
        FIRMWARE_IS_FOUND=1
        FIRMWARE_INSTALL_REQUIRED=1
      fi
    fi
  done

  if [ ! -f "./version" ]; then
    echo "[Error] ./version doesn't exist!" > /dev/console
    process_exit 1
  fi

  if [ "$FIRMWARE_IS_FOUND" -eq "0" ] && [ "$BOOTLOADER_IS_FOUND" -eq "0" ]; then
    echo "[Error] archive content isn't correct!" > /dev/console
    process_exit 1
  fi
}

check_actual_version() {
	if [ "$FIRMWARE_IS_FOUND" -eq "1" ] ; then
		cmp -s version /etc/version
	
		if [ "$?" == 0 ]; then
			echo "The actual firmware version is already installed" > /dev/console
			FIRMWARE_INSTALL_REQUIRED=0
		fi
	fi
	
	if [ "$BOOTLOADER_IS_FOUND" -eq "1" ] ; then
		ubootenv -g ver | cut -d '=' -f 2 > $BOOT_VERSION_FILE
		cmp -s version $BOOT_VERSION_FILE

		if [ "$?" == 0 ]; then
			echo "The actual bootloader version is already installed" > /dev/console
			BOOTLOADER_INSTALL_REQUIRED=0
		fi

		rm $BOOT_VERSION_FILE
	fi
}

handle_version() {
	if [[ $CHECK_VERSION -ne 0 ]]; then
	  echo "Check actual firmware version" > /dev/console
	  check_actual_version
	fi
}

check_upgrade_needing() {
  handle_version

  if [ "$FIRMWARE_IS_FOUND" -eq "1" ] && [ "$BOOTLOADER_IS_FOUND" -eq "0" ]; then
    if [ "$FIRMWARE_INSTALL_REQUIRED" -eq "0" ]; then
      process_exit 2
    fi
  fi

  if [ "$FIRMWARE_IS_FOUND" -eq "0" ] && [ "$BOOTLOADER_IS_FOUND" -eq "1" ]; then
    if [ "$BOOTLOADER_INSTALL_REQUIRED" -eq "0" ]; then
      process_exit 2
    fi
  fi

  if [ "$FIRMWARE_IS_FOUND" -eq "1" ] && [ "$BOOTLOADER_IS_FOUND" -eq "1" ]; then
    if [ "$FIRMWARE_INSTALL_REQUIRED" -eq "0" ] && [ "$BOOTLOADER_INSTALL_REQUIRED" -eq "0" ]; then
      process_exit 2
    fi
  fi

  VERSION="$(cat ./version)"
  echo "Version $VERSION will be installed" > /dev/console
}

upgrade_firmware() {
	vdimg=$(get_vol_num_from_vol_name "0" "${VOLUME_NAME__DTB}${BANK_NEXT}")
	vkimg=$(get_vol_num_from_vol_name "0" "${VOLUME_NAME__KERNEL}${BANK_NEXT}")
	vrimg=$(get_vol_num_from_vol_name "0" "${VOLUME_NAME__ROOTFS}${BANK_NEXT}")

	CUR_STAGE=$(($CUR_STAGE + 1))
	echo "  Stage ${CUR_STAGE}: updating $vdimg" > /dev/console
	ubiupdatevol "$vdimg" "$dtb_file.bin"

	if [ "$?" -ne "0" ]; then
		echo "[Error] ubiupdatevol $vdimg $dtb_file.bin" > /dev/console
		process_exit 8
	fi

	CUR_STAGE=$(($CUR_STAGE + 1))
	echo "  Stage ${CUR_STAGE}: updating $vkimg" > /dev/console
	ubiupdatevol "$vkimg" "$kernel_file.bin"

	if [ "$?" -ne "0" ]; then
		echo "[Error] ubiupdatevol $vkimg $kernel_file.bin" > /dev/console
		process_exit 8
	fi

	CUR_STAGE=$(($CUR_STAGE + 1))
	echo "  Stage ${CUR_STAGE}: updating $vrimg" > /dev/console
	ubiupdatevol "$vrimg" "$rootfs_file.bin"

	if [ "$?" -ne "0" ]; then
		echo "[Error] ubiupdatevol $vrimg $rootfs_file.bin" > /dev/console
		process_exit 8
	fi
}

upgrade_bootloader() {
  for i in $bootloader_mtd_list; do
    if [ -f "./$i.bin" ]; then
      echo "./$i.bin" > /dev/console

      if [ "$i" == "$fip_file" ]; then
        #TODO: we need to check FIP_SIZE_LIMIT=1048576
        flash_erase ${DEVICE_BOOT} ${FIP1_OFFSET} ${FIP_SIZE_IN_BLOCKS} > /dev/null
        nandwrite -s ${FIP1_OFFSET} -p ${DEVICE_BOOT} "./$i.bin" > /dev/null

        flash_erase ${DEVICE_BOOT} ${FIP2_OFFSET} ${FIP_SIZE_IN_BLOCKS} > /dev/null
        nandwrite -s ${FIP2_OFFSET} -p ${DEVICE_BOOT} "./$i.bin" > /dev/null

        log_upgrade_ok ${DEVICE_BOOT} 0
      fi

      if [ "$i" == "$env_file" ]; then
        flash_erase ${DEVICE_ENV} ${ENV1_OFFSET} ${ENV_SIZE_IN_BLOCKS} > /dev/null
        nandwrite -s ${ENV1_OFFSET} -p ${DEVICE_ENV} "./$i.bin" > /dev/null

        flash_erase ${DEVICE_ENV} ${ENV2_OFFSET} ${ENV_SIZE_IN_BLOCKS} > /dev/null
        nandwrite -s ${ENV2_OFFSET} -p ${DEVICE_ENV} "./$i.bin" > /dev/null

        log_upgrade_ok ${DEVICE_ENV} 0
      fi
    fi
  done
}

upgrade_process() {
  #Step 1. Checking
  check_files_exist
  check_upgrade_needing

  #Step 2. Send upgrade notification
  type send-upgrade-notification >/dev/null

  if [ "$?" -eq "0" ]; then
    send-upgrade-notification
    NOTIFY_STATUS=$?

    if [ "$NOTIFY_STATUS" -eq "5" ]; then
      echo "There are active calls on phone, upgrade aborted" > /dev/console
      process_exit 5
    fi
  fi

  #Step 3. Start the upgrade process
  read_env_variables

  BANK_1_TRY=3
  BANK_2_TRY=3

  save_env_variables

  if [ "$BOOTLOADER_INSTALL_REQUIRED" -eq "1" ]; then
    echo "===== Start of BOOTLOADER upgrade =====" > /dev/console

    upgrade_bootloader

    echo "Recovering variables" > /dev/console

    save_env_variables

    echo "===== End of BOOTLOADER upgrade =====" > /dev/console
  fi

  if [ "$FIRMWARE_INSTALL_REQUIRED" -eq "1" ]; then
    echo "===== Start of FIRMWARE upgrade =====" > /dev/console

    if [ "$BANK_ACTIVE" -eq "1" ]; then
      BANK_NEXT=2
    else
      BANK_NEXT=1
    fi

    upgrade_firmware

    if [ "$BANK_ACTIVE" -eq "1" ]; then
      BANK_ACTIVE=2
    else
      BANK_ACTIVE=1
    fi
    save_env_variables

    echo "===== End of FIRMWARE upgrade =====" > /dev/console
  fi

  echo "Upgrade has been successfully completed. Well done!" > /dev/console
}

###############################################################################################

if [ -f ${FILE_LOCK_UPGRADE} ]; then
  echo "[Warning] You need to reboot to complete the upgrade process!" > /dev/console
  exit 1
fi

while getopts nv flag
do
    case "${flag}" in
        n) NO_REBOOT="1";;
        v) CHECK_VERSION="1";;
    esac
done

echo "1" > ${FILE_LOCK_UPGRADE}

upgrade_process

if [[ $NO_REBOOT -ne 1 ]]; then
  echo "The system will be rebooted" > /dev/console
  reboot
fi
